(require deftemplates)
(require defqueries)

(set-current-module CONTENT)

(defrule print-module-content
    (declare (salience 100))
    (MAIN::model (tick ?tick))
    =>
    ((System.out) println (str-cat "MODULE CONTENT started")))


;; ----- endorsement related rules --------------------------

(defrule similarity-endorsement
    "endorse most similar and other similar"
;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::person (name ?person))
    (exists (MAIN::similarity-index (clauseOwner ?person)))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule similarity-endorsement for " ?person)))
    (bind ?indices (create$))
    (bind ?others (create$))
    (bind ?qr (run-query* MAIN::similarity-index-query ?person ?tick))
    (while (?qr next)
        (bind ?indices (create$ ?indices (?qr get similarity)))
        (bind ?others (create$ ?others (?qr get other-person))))
;    (watch facts)
    (for (bind ?i 1)(<= ?i (length$ ?others))(++ ?i)
        (bind ?index (nth$ ?i ?indices))
        (if (> ?index 0) then
            (assert
                (endorsed 
                    (endorsed-object (nth$ ?i ?others))
                    (endorsement similar)
                    (uniquifiers (create$ ?index))
                    (clauseOwner ?person)
                    (tickStamp ?tick)))))
    
;    (unwatch facts)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
)

(defrule most-similar-endorsement
	"endorse others with the highest similarity indes"
;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::person (name ?person))
    (exists (MAIN::endorsed (endorsement similar)(clauseOwner ?person)))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule most-similar-endorsement for " ?person)))
    (bind ?max 0)
    (bind ?others (create$))
    (bind ?qr (run-query* MAIN::specific-endorsement-query ?person similar ?tick))
    (while (?qr next)
        (bind ?current (nth$ 1(?qr get uniqs)))
        (if (> ?current ?max) then
            (bind ?others (create$ (?qr get endorsee)))
            (bind ?max ?current)
            else
            (if (= ?max ?current) then
                (bind ?others (create$ ?others (?qr get endorsee))))))
;    (watch facts)
    (foreach ?other ?others
        (assert
	        (endorsed
	            (endorsed-object ?other)
	            (endorsement most-similar)
	            (clauseOwner ?person)
	            (tickStamp ?tick))))
;    (unwatch facts)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
)

;; ----- household economy related rules --------------------

(defrule borrow-food 
    "let household ask relatives/neighbours for help if there's not enough money to buy all the food needed"
    (MAIN::model (week ?week)(OBJECT ?m-obj))
    (help-needed (household ?hh)(amount ?amount)(tickStamp ?tick))
    ?f <- (asked-for-help (requester ?hh)(hh-list $?helper-list){tickStamp <= ?tick})
    (MAIN::household (id ?hh)(headID ?head))
    (MAIN::household (id ?other-hh &: (not (member$ ?other-hh ?helper-list)))(headID ?other-head)
        (monthlyFoodCost ?cost){expectedIncome > ?amount}
        /*(spendingCommitments ?commitments){cash >= (+ ?amount ?commitments)}*/)
    (endorsed (endorsement is-kin|is-neighbour)(clauseOwner ?head)(endorsed-object ?other-head))
    =>
	(if (> (+ ?tick ?week) 350) then
        ((System.out) println (str-cat "rule borrow-food for " ?hh)))
        
    ;; ask other hh for help with a chance of 60 to 90% (increasing with time)
    (if (call ?m-obj probabilityp (+ 0.6 (* 0.1 ?week))) then 
	    (assert
	       	(MAIN::message
	       		(header borrow-food-request)
	    		(recipient ?other-hh)
	    		(sender ?hh)
	    		(tickStamp (+ ?tick ?week))
	    		(slot-names amount)
	    		(slot-values ?amount)))
	    (watch facts)
	    (modify ?f (hh-list (create$ ?helper-list ?other-hh)))
	    (unwatch facts)
	    (dump (str-cat ?hh " attempts to borrow food for " ?amount " Rand from " ?other-hh))
    )
    
	(if (> (+ ?tick ?week) 350) then
        ((System.out) println "end of rule"))      
)


;; -------- burial society related rules --------------------

(defrule create-burial-society
	"create a new burial society from a group of friends"
    (MAIN::model (OBJECT ?m-obj)(tick ?tick))
    ?f <- (MAIN::group (group-id ?leader) (group-type burial-society) (members $?members))
	=>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule create-burial-society for " ?leader)))
    ;; fetch Java objects of members
    (bind ?list (new java.util.ArrayList))
    (foreach ?member ?members
		(bind ?qr (run-query* MAIN::person-object-query ?member))
        (while (?qr next)
        	(?list add (?qr get object)))
     )    
    ;; create burial society as new shadow fact
    (new BurialSociety 0 ?list ?m-obj) ;; will be put into Jess from Java (in constructor of BurialSociety)
    (retract ?f)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)        

(defrule join-burial-society
   "if there is enough cash at the end of the month (TODO: over several months) and no-one in the household is yet a member of a b.s."
    (MAIN::model (week 2)(tick ?tick))
    (MAIN::household (id ?hh) (cash ?cash &: (>= ?cash 50)) (village ?village))
    (MAIN::person (name ?name) (householdID ?hh) (head TRUE) (burialSocieties $?bs-list &: (= (length$ ?bs-list) 0)) (OBJECT ?p-obj))
    (not (MAIN::death (household ?hh)))
    =>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule join-burial-society for " ?hh)))
    ;; which one to join though?
    ;; could do a query on which b.s my friends belong to
    ;; then pick one of them (the cheapest? the one where most of my friends are?) -- important: has to be in the village!
	(bind ?bs-list (create$))
    (bind ?qr (run-query* MAIN::burial-societies-query ?name ?village))
	(while (?qr next)
		(bind ?bs-list (create$ ?bs-list (?qr get friend-bs)))
    )
;	(printout t ?name " retrieves the following list of friends' burial societies: " ?bs-list  crlf)
    (if (<= (length$ ?bs-list) 0) then
        ;; no burial societies to join --> try to start a new one
;        (watch facts)
    	(assert 
        	(MAIN::group-wanted
        		(type burial-society)
    			(person ?name)
    			(tickStamp (current-tick))))
;		(unwatch facts)
    else 
		;; for now: pick the cheapest
		(bind ?min 1000)
		(bind ?my-bs nil)
		(bind ?my-bs-obj nil)
		(foreach ?bs ?bs-list
	    	(bind ?qr (run-query* MAIN::specific-burial-society-query ?bs))
	    	(while (?qr next)
				(bind ?fee (?qr get fee))
				(if (< ?fee ?min) then
		    		(bind ?min ?fee)
					(bind ?my-bs ?bs)
					(bind ?my-bs-obj (?qr get obj))
				)
			)
		)
    	;; join the society!
		((System.out) println (str-cat ?name " joins " ?my-bs))
        (dump (str-cat ?name " joins " ?my-bs))
        (if (eq ?my-bs nil) then
            (dump (str-cat "RULE FAILED to find the cheapest burial society from this list: " ?bs-list))
         else 
			(call ?my-bs-obj addMember ?p-obj))
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)
   
(defrule leave-burial-society-after-death
    "dead member leaves burial society; leaves a reminder to hh which burial society he/she was ins"
	(MAIN::death (household ?hh) (name ?deceased) (tickStamp ?tick))
    (MAIN::burialSociety (name ?bs) (memberIDs $?members &: (member$ ?deceased ?members)) (OBJECT ?bs-obj))
    (MAIN::person (name ?deceased) (OBJECT ?deceased-obj))
    =>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule leave-burial-society-after-death for " ?deceased " of " ?hh)))
    ;; remove dead member from burial society    
    (call ?bs-obj removeMember ?deceased-obj)  
    ;; leave a reminder to hh as to which burial society's membership they might want to continue    
    (assert
        (MAIN::membership-reminder
            (deceased ?deceased)
            (household ?hh)
            (group ?bs)
            (type burial-society)
            (tickStamp ?tick)))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
    
)


;; --------- employment related rules ------------------------

(defrule occasional-employment
	"if sufficiently rich, hire one person with probability of 0.15 in any week"
    (declare (no-loop TRUE))
;	(MAIN::model (OBJECT ?m-obj)(standardWage ?w))
	(MAIN::household (id ?employer) (cash ?cash) (spendingCommitments ?commitments) {expectedIncome > 500} (model ?m-obj))
    (not (MAIN::marriage (groom-household ?employer))) ;; being able to pay lobola has priority
    (not (MAIN::employees-demanded (employer ?employer)))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule occasional-employment for " ?employer)))
    (bind ?available-for-hiring (- ?cash ?commitments))
    (bind ?max-employees (div ?available-for-hiring ?*standard-wage*))
    (bind ?demand 0)
    (for (bind ?i 0)(< ?i ?max-employees)(++ ?i)
        (if (?m-obj probabilityp 0.15) then
            (++ ?demand)))
    (if (> ?demand 0) then
;		((System.out) println (str-cat ?employer " with cash " ?cash " and commitments " ?commitments  " advertises " ?demand " piece jobs (out of "  ?max-employees " possible)"))
		(dump (str-cat ?employer " with cash " ?cash " and commitments " ?commitments  " advertises " ?demand " piece jobs (out of "  ?max-employees " possible)"))
;       (watch facts)
	    (assert
	        (employees-demanded
	            (employer ?employer)
	            (number-demanded ?demand)
	            (tickStamp ?tick)))
;	    (unwatch facts)
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule assess-employment
    "have employer rate the performance of the employee"
    (MAIN::model (OBJECT ?m-obj)(tick ?tick))
    (MAIN::employment (employee ?employee) (employer ?employer &: (neq ?employer government)) (end-date ?tick))
    (MAIN::person (name ?hh-head) (head TRUE) (householdID ?employer) (tag $?employer-tag))
	(MAIN::person (name ?employee) (tag $?employee-tag))
	?f <- (MAIN::job-performance (employee ?employee) (employer ?employer) (reliability ?r) (capability ?c) (num-jobs ?num-jobs){tickStamp < ?tick})
	=>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule assess-employment for " ?employer " and employee " ?employee)))
   	;; performance of a person is based on their tags
    ;; no, it's now randomly assigned
	(bind ?reliability 0)
;	(if (>= (nth$ 1 $?employee-tag) (nth$ 1 $?employer-tag)) then
    (if (call ?m-obj probabilityp 0.7) then  
    	(bind ?reliability 1)
    )
	(bind ?capability 0)
;	(if (>= (nth$ 2 $?employee-tag) (nth$ 2 $?employer-tag)) then
    (if (call ?m-obj probabilityp 0.8) then
    	(bind ?capability 1)
	)
	(modify ?f 
    	(reliability (+ ?r ?reliability)) 
    	(capability (+ ?c ?capability))
		(num-jobs (+ ?num-jobs 1))
		(tickStamp ?tick))
;    ((System.out) println (str-cat ?employer " rates job performance of " ?employee " as " ?reliability " / " ?capability))
    (dump (str-cat ?employer " rates job performance of " ?employee " as " ?reliability " / " ?capability))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule assess-first-employment
    "have employer rate the performance of the employee"
    (MAIN::model (OBJECT ?m-obj)(tick ?tick))
    (MAIN::employment (employee ?employee) (employer ?employer &: (neq ?employer government)) (end-date ?tick))
    (MAIN::person (name ?hh-head) (head TRUE) (householdID ?employer) (tag $?employer-tag))
	(MAIN::person (name ?employee) (tag $?employee-tag))
	(not (MAIN::job-performance (employee ?employee) (employer ?employer)))
	=>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule assess-first-employment for " ?employer " and employee " ?employee)))
    ;; performance of a person is based on their tags
	(bind ?reliability 0)
;	(if (>= (nth$ 1 $?employee-tag) (nth$ 1 $?employer-tag)) then
    (if (call ?m-obj probabilityp 0.7) then  
    	(bind ?reliability 1)
    )
	(bind ?capability 0)
;	(if (>= (nth$ 2 $?employee-tag) (nth$ 2 $?employer-tag)) then
    (if (call ?m-obj probabilityp 0.8) then
    	(bind ?capability 1)
	)
	(assert 
    	(MAIN::job-performance
    		(employee ?employee)
			(employer ?employer)
			(reliability ?reliability)
			(capability ?capability)
			(num-jobs 1)
			(tickStamp ?tick)))
;    ((System.out) println (str-cat ?employer " rates job performance of " ?employee " as " ?reliability " / " ?capability ))
    (dump (str-cat ?employer " rates job performance of " ?employee " as " ?reliability " / " ?capability ))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

;; ----------- migration related rules -----------------------

(defrule send-remittance
	"a migrant worker sends remittance home"
	(MAIN::model (week 2))
	(MAIN::person (alive TRUE) (name ?migrant) (householdID ?hh) (migrant TRUE) (atSchool FALSE) (model ?m-obj))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule send-remittance for " ?migrant " of " ?hh)))
        
	;; for a first try just do it randomly
	(if (call ?m-obj probabilityp ?*remittance-probability*) then
        (bind ?amount (call ?m-obj randomInt ?*min-remittance* ?*max-remittance*)) ;; let amount be influenced by education level of migrant?
    	(assert 
    		(MAIN::payment
    			(payer ?migrant)
				(recipient ?hh)
				(cause remittance)
				(amount ?amount)
				(payment-date ?tick)))
        ;((System.out) println (str-cat ?migrant " sends " ?amount " remittance home to " ?hh ))
        (dump (str-cat ?migrant " sends " ?amount " remittance home to " ?hh ))
	)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)
            
(defrule migrate-for-work
	"migrate to find a job if out of school with no employment and hh needs money"
    (MAIN::model (OBJECT ?m-obj)(tick ?tick))
    ?h <- (MAIN::household (id ?hh) (expectedIncome ?income) {monthlyFoodCost >= ?income} (cash ?cash) (spendingCommitments ?commitments))
    (or ?p <- (MAIN::person (alive TRUE)(householdID ?hh)(head FALSE){age > 15}(name ?name) (migrant FALSE)(atSchool FALSE)(onVisit FALSE))
        ?p <- (MAIN::person (alive TRUE)(householdID ?hh)(head TRUE)(gender male)(name ?name)(migrant FALSE)(atSchool FALSE)(onVisit FALSE)))
    (not (MAIN::employment (employee ?name) {end-date > ?tick}))
    (not (MAIN::payment (payer ?hh)(cause migration)(payment-date ?tick))) ;; only one member of hh per tick
	=>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule migrate-for-work for " ?name " of " ?hh)))
     ;; should decision depend on education? ##
    (dump (str-cat ?name " is contemplating migration"))             
	;; (a) have to find already migrated relatives or friends to stay with
	;; find migrant relatives
	(bind ?migrants (create$))
	(bind ?qr (run-query* MAIN::migrant-relatives-query ?name ?hh))
	(while (?qr next)
    	(bind ?migrants (create$ ?migrants (?qr get relative)))
    )
	(if (= (length$ ?migrants) 0) then
        ((System.out) println (str-cat "no migrant relatives found for " ?name " of " ?hh))
    	;; find migrant friends
		(bind ?qr (run-query* MAIN::migrant-friends-query ?name))
		(while (?qr next)
    		(bind ?migrants (create$ ?migrants (?qr get friend)))
		)
	)
    ;; (b) need to have enough money for the journey
    (if (> (length$ ?migrants) 0) then
        ((System.out) println (str-cat ?name " has migrant friends and/or relatives"))
        (bind ?costs (call ?m-obj randomInt ?*min-travel-costs* ?*max-travel-costs*)) ;; how much are travel costs??
    	(if (> (- (+ ?cash ?income) ?commitments) ?costs) then
        	;; go away to find a job
        	(modify ?p (migrant TRUE))
        	(modify ?h (spendingCommitments (+ ?commitments ?costs)))
        	(assert
            	(MAIN::payment
            		(amount ?costs)
        			(cause migration)
        			(payer ?hh)
        			(payment-date ?tick)))
;        	((System.out) println (str-cat ?name " of " ?hh " migrates to find a job"))
            (dump (str-cat ?name " of " ?hh " migrates to find a job"))
        )
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule visit-home-for-summer-holidays
    "migrant school/college attendants will come home for the holidays"
    (MAIN::model (week 0)(month 11)(OBJECT ?m-obj))
    ?p <- (MAIN::person  (alive TRUE)(name ?person)(migrant TRUE)(atSchool TRUE)(householdID ?hh))
    ?h <- (MAIN::household (id ?hh)(spendingCommitments ?commitments))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule visit-home-for-summer-holidays for " ?person " of " ?hh)))
    (modify ?p (migrant FALSE)(onVisit TRUE)(lastVisit (+ ?tick 6))) ;; summer holidays last 6 weeks
    (bind ?travel-costs (call ?m-obj randomInt ?*min-travel-costs* ?*max-travel-costs*))
    (modify ?h (spendingCommitments (+ ?commitments ?travel-costs)))
    (assert
        (MAIN::payment
            (amount ?travel-costs)
            (cause travel)
            (payer ?hh)
            (payment-date ?tick)))
    (dump (str-cat "Migrant "?person " visits home (" ?hh ") for the summer holidays"))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule visit-home-for-term-holidays
    "migrant school/college attendants will come home for the holidays if the household can afford it"
    (MAIN::model (week 3) (month 2 | 5)(OBJECT ?m-obj))
    ?p <- (MAIN::person  (alive TRUE)(name ?person)(migrant TRUE)(atSchool TRUE)(householdID ?hh))
    ?h <- (MAIN::household (id ?hh)(cash ?cash)(spendingCommitments ?commitments)(expectedIncome ?income))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule visit-home-for-term-holidays for " ?person " of " ?hh)))
        (bind ?travel-costs (call ?m-obj randomInt ?*min-travel-costs* ?*max-travel-costs*))
    (if (> (+ ?cash ?income) ?travel-costs) then
        (modify ?p (migrant FALSE)(onVisit TRUE)(lastVisit (+ ?tick 3))) ;; holidays last 3 weeks
        (modify ?h (spendingCommitments (+ ?commitments ?travel-costs)))
        (assert
            (MAIN::payment
                (amount ?travel-costs)
                (cause travel)
                (payer ?hh)
                (payment-date ?tick)))
        (dump (str-cat "Migrant " ?person " visits home (" ?hh ") for the term holidays"))
    else
        (dump (str-cat "Migrant " ?person " can't afford to visit home (" ?hh ") for the term holidays"))
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule visit-home
    "every now and then a migrant will visit home for up to 2 months"
    (MAIN::model (tick ?tick)(OBJECT ?m-obj))
    ?p <- (MAIN::person  (alive TRUE)(name ?person)(migrant TRUE)(atSchool FALSE)(lastVisit ?last &: (< ?last (- ?tick 24)))(householdID ?hh))
    =>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule visit-home for " ?person " of " ?hh)))
    (bind ?weeks (call ?m-obj randomInt 0 8)) ;; might not visit
    (if (> ?weeks 0) then 
    	(modify ?p (migrant FALSE)(onVisit TRUE)(lastVisit (+ ?tick ?weeks)))
    	(dump (str-cat "Migrant "?person " visits home (" ?hh ") for " ?weeks " weeks."))
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule resume-migration
    "return after visit home"
    ?p <- (MAIN::person  (alive TRUE)(name ?person)(migrant FALSE)(onVisit TRUE)(atSchool ?at-school)(lastVisit ?last)(currentPartners $?partners)(OBJECT ?p-obj))
    (MAIN::model (tick ?last))
    =>
	(if (> ?last 350) then
        ((System.out) println (str-cat "rule resume-migration for " ?person " of " ?hh)))
    (modify ?p (migrant TRUE)(onVisit FALSE))
    (if (eq ?at-school TRUE) then
        (dump (str-cat "Migrant " ?person " returns to school/college"))
     else
        (dump (str-cat "Migrant " ?person " returns to work"))   
    )
	(if (> ?last 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule finish-migration-due-to-marriage
    "return home due to marriage"
    (MAIN::marriage (bride ?bride)(groom ?groom)(tickStamp ?tick))
    ?p <- (MAIN::person (name ?bride|?groom)(migrant TRUE))
    =>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule finish-migration-due-to-marriage for " (fact-slot-value ?p name) )))
    (modify ?p (migrant FALSE))
    (dump (str-cat (fact-slot-value ?p name) " stops migration to get married in the village"))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule finish-migration-due-to-old-age
    "migrants older than 55 tend to return home to their village"
    ?p <- (MAIN::person (name ?person)(migrant TRUE){age > 55})
    (MAIN::model (month 5|6|7|11)(OBJECT ?m-obj))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule finish-migration-due-to-old-age for " ?person )))
           
    (if (call ?m-obj probabilityp (* 0.1 (abs (- 55 ?age)))) then
        (modify ?p (migrant FALSE))
        (dump (str-cat ?person " stops migration at age " ?age))
    ) 
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
          
)

(defrule finish-migration-due-to-job-loss
    "return home when losing the migrant job"
    ?p <- (MAIN::person (name ?person)(migrant TRUE)(achievedEducation ?education))
    (MAIN::model (month 5|6|7|11)(OBJECT ?m-obj))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule finish-migration-due-to-job-loss for " ?person )))
        ;; higher probability when education level is low?
 	(if (call ?m-obj probabilityp 0.01) then
        (modify ?p (migrant FALSE))
        (dump (str-cat ?person " stops migration due to job loss"))
    ) 
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
          
)


;; --------------- education related rules ----------------

(defrule finish-secondary-school
    "at the beginning of a school year a pupil > 15 years checks if he/she can continue their education"
	(MAIN::model (week 1) (month 0))
	?p <- (MAIN::person (name ?name) (householdID ?hh)(atSchool TRUE) (achievedEducation some-secondary)(age ?age &: (> ?age 15)))
	?h <- (MAIN::household (id ?hh) (cash ?cash) (expectedIncome ?income) (spendingCommitments ?commitments))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule finish-secondary-school for " ?name " of " ?hh )))
    
    ;	(bind ?fee 200) ;; how high are school fees???
	(if (< (+ ?cash (- ?income ?commitments)) ?*school-fee*) then	
    	;; not enough money
		(modify ?p (atSchool FALSE))
;		((System.out) println (str-cat ?name " of " ?hh " stops school at age " ?age " due to lack of money"))
        (dump (str-cat ?name " of " ?hh " stops school at age " ?age " due to lack of money"))
	else
		;; stay at school
		(modify ?h (spendingCommitments (+ ?commitments ?*school-fee*)))
		(assert
    		(MAIN::payment
    			(amount ?*school-fee*)
				(payer ?hh)
				(cause school-fee)
				(payment-date (current-tick))))
		
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule start-university
    "at the beginning of a school year a pupil with finished secondary education checks if they can attend college/uni"
	(MAIN::model (week 1) (month 0))
	?p <- (MAIN::person  (alive TRUE)(name ?name) (householdID ?hh) (maritalStatus single) (achievedEducation secondary) (age ?age &: (< ?age 21)) (OBJECT ?p-obj))
	?h <- (MAIN::household (id ?hh) (cash ?cash) (expectedIncome ?income) (spendingCommitments ?commitments))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule start-university for " ?name " of " ?hh )))
    ;	(bind ?fee 10000) ;; how high are uni fees?
	;; (a) need to find a relative or friend with whom to stay when going to college/uni
	;; find migrant relatives
	(bind ?migrants (create$))
	(bind ?qr (run-query* MAIN::migrant-relatives-query ?name ?hh))
	(while (?qr next)
    	(bind ?migrants (create$ ?migrants (?qr get relative)))
    )
	(if (= (length$ ?migrants) 0) then
    	;; find migrant friends
		(bind ?qr (run-query* MAIN::migrant-friends-query ?name))
		(while (?qr next)
    		(bind ?migrants (create$ ?migrants (?qr get friend)))
		)
	)
	(if (> (length$ ?migrants) 0) then 
		;; (b) need to have enough money to pay the college fees (and the transport to the college town)
		(if (> (- (+ ?cash ?income) ?commitments) ?*college-fee*) then
	    	;; off to college we go!
			(modify ?p (migrant TRUE) (atSchool TRUE)) ;;### why does (achievedEducation tertiary) not work here???
			(call ?p-obj setAchievedEducation "tertiary")
			(modify ?h (spendingCommitments (+ ?commitments ?*college-fee*)))
			(assert
	    		(MAIN::payment
	    			(amount ?*college-fee*)
					(payer ?hh)
					(cause school-fee)
					(payment-date (current-tick))))
;			((System.out) println (str-cat ?name " of " ?hh " goes away to college/uni"))
            (dump (str-cat ?name " of " ?hh " goes away to college/uni"))
    	)
	;; if not enough money, could ask a relative to pay??
	;; or start saving up in a stokvel so will be able to go a year later??
	)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule pay-college-fees
    "pay fees at beginning of uni year"
    (MAIN::model (week 1) (month 0))
    ?p <- (MAIN::person (name ?name)(atSchool TRUE)(achievedEducation tertiary)(householdID ?hh))
    ?h <- (MAIN::household (id ?hh)(spendingCommitments ?commitments)(cash ?cash)(expectedIncome ?income))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule pay-college-fees for " ?name " of " ?hh )))
    ;	(bind ?fee 10000) ;; how high are uni fees?
	(if (> (- (+ ?cash ?income) ?commitments) ?*college-fee*) then
		(modify ?h (spendingCommitments (+ ?commitments ?*college-fee*)))
		(assert
	    	(MAIN::payment
	    		(amount ?*college-fee*)
				(payer ?hh)
				(cause school-fee)
				(payment-date (current-tick))))
        (dump (str-cat ?hh " pays college fees for " ?name))
    else
        (dump (str-cat ?hh " can't afford college fees for " ?name))
        ;; now who pays the fees?? Or does the person have to stop their college educatiom and come home??
     	;; for now: let them come home
        (modify ?p (atSchool FALSE)(migrant FALSE)(onVisit FALSE))
    )   
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
                       
)

;; ---------------- marriage related rules ------------------

(defrule find-a-suitable-partner
    "if a man has enough money to pay the lobola he's ready to get married"
;	(MAIN::model (OBJECT ?m-obj))
	(MAIN::person  (alive TRUE)(name ?name) (householdID ?hh) (maritalStatus single|widowed) (gender male) 
        (currentPartners $?partners &: (> (length$ ?partners) 0))
        (atSchool FALSE)(age ?age &: (> ?age 24)) (agentEndorsementScheme ?end-scheme)(model ?m-obj))
	(MAIN::household (id ?hh) (spendingCommitments ?commitments)(cash ?cash &: (> (- ?cash ?commitments) ?*min-lobola*)))
	(not (marriage (groom-household ?hh))) ;; don't have two males from the same household marry at the same time
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule find-a-suitable-partner for " ?name " of " ?hh )))
    ;    (bind ?list (create$))
;	(bind ?qr (run-query* MAIN::suitable-partner-query ?hh ?age))
;	(while (?qr next)
;    	(bind ?list (create$ ?list (?qr get name)))
;    )
;	(if (> (length$ ?list) 0) then
        (bind ?list ?partners) ;; <--- only choose from current sexual partners
     	;; use endorsements to evaluate each possible partner
        (bind ?end-values (create$))
        (foreach ?partner ?list
        	(bind ?qr (run-query* MAIN::endorsements-query ?name ?partner ?tick))
            (bind ?endorsements (create$))
            (while (?qr next)
            	(bind ?endorsements (create$ ?endorsements (?qr get endorsement)))
            )
        	(bind ?end-values (create$ ?end-values (endorsement-value ?end-scheme ?endorsements)))
        )
        ;; sort list of endorsement values in descending order of values
        (bind ?indices (sort-values ?end-values TRUE))
        (bind ?my-partner (nth$ (nth$ 1 ?indices) ?list))
        ;(dump (str-cat "chosen partner is one of current partners: " (member$ ?my-partner ?partners)))
    	(if (call ?m-obj fitForMarriage ?name ?my-partner) then
			;; announce marriage
			(bind ?qr (run-query* MAIN::person-household-query ?my-partner))
			(while (?qr next)
		    	(bind ?partner-household (?qr get hh)))
			(assert
		    	(MAIN::marriage
		    		(bride ?my-partner)
					(bride-household ?partner-household)
					(groom ?name)
					(groom-household ?hh)
					(tickStamp ?tick)))
			;; perform marriage
			(call ?m-obj marry ?name ?my-partner)
    	)
;   	)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule build-a-house
	"if there's enough money the newly wed couple move out to their own household (only if groom isn't head of hh already)"
;	(declare (no-loop TRUE))
;	(MAIN::model (OBJECT ?m-obj))
	?u <- (MAIN::upcoming-household (person ?groom) (tickStamp ?t)) ;; still living at parents'
	(MAIN::person  (alive TRUE)(name ?groom) (head FALSE)(householdID ?hh) (OBJECT ?groom-obj)(model ?m-obj)) 
	?h <- (MAIN::household (id ?hh) (spendingCommitments ?commitments) (expectedIncome ?income){cash > ?*house-cost*})
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule build-a-house for " ?groom " of " ?hh )))
    ;; how much does a house cost??
    (bind ?margin (integer (* ?*house-cost* 0.2)))
	(bind ?amount (call ?m-obj randomInt (- ?*house-cost* ?margin) (+ ?*house-cost* ?margin)))
    (if (> (- (+ ?cash ?income) ?commitments) ?amount) then
    	;; build a house and move out
		(bind ?new-house (call ?m-obj buildNewHouse ?groom-obj))  ;; this also puts the new household as a shadow fact into Jess
		(modify ?h (spendingCommitments (+ ?commitments ?amount)))
        (bind ?tick (current-tick))
		(assert 
    		(MAIN::payment
				(amount ?amount)
				(cause new-house)
				(payer ?hh)
				(recipient ?new-house)
				(payment-date ?tick)))
        (retract ?u)
		((System.out) println (str-cat ?groom " moves out from " ?hh " into new " ?new-house " with his spouse (and children if any), " (- ?tick ?t) " ticks after marriage"))
        (dump (str-cat ?groom " moves out from " ?hh " into new " ?new-house " with his spouse (and children if any), " (- ?tick ?t) " ticks after marriage"))
 	)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

;; --------------- stokvel related rules ------------------

(defrule save-up-for-lobola
	"a single male adult with a job will try to join/form a stokvel to save for lobola"
    (MAIN::person  (alive TRUE)(name ?name) (gender male) (maritalStatus single) {age > 21} (householdID ?hh)
        (stokvels $?s &: (<= (length$ ?s) 0)) (OBJECT ?p-obj) (model ?m-obj))
    (MAIN::employment (employee ?name))
    (MAIN::household (id ?hh) (village ?village))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule save-up-for-lobola for " ?name " of " ?hh )))
    ;; look for existing stokvels any friends belong to
	(bind ?s-list (create$))
    (bind ?qr (run-query* MAIN::stokvels-query ?name ?village))
	(while (?qr next)
		(bind ?s-list (create$ ?s-list (?qr get friend-st)))
    )
    (if (<= (length$ ?s-list) 0) then
        ;; try to form a new stokvel
        (assert 
        (MAIN::group-wanted
            (person ?name)
            (type stokvel)
            (tickStamp (current-tick))))
    else
        ;; join the stokvel where most of my friends are 
        ;; --> that's the stokvel featuring prominently in the list
        (bind ?my-stokvel (element-with-highest-occurrence ?s-list))
        (bind ?my-stokvel-obj nil)
        (bind ?qr (run-query* MAIN::specific-stokvel-query ?my-stokvel))
        (while (?qr next) (bind ?my-stokvel-obj (?qr get obj)))
        (dump (str-cat ?name " joins " ?my-stokvel))
        (call ?my-stokvel-obj addMember ?p-obj)
        (call ?m-obj recordStokvelEvent ?my-stokvel 
            (str-cat ?name " (" ?age ") chooses " ?my-stokvel " from " ?s-list " to join to save up for lobola") ?p-obj)
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule save-up-for-house
	"a married male adult with a job will try to form/join a stokvel to save for building a house"
    (MAIN::person  (alive TRUE)(name ?name)(gender male)(maritalStatus married)(householdID ?hh)
        (stokvels $?s &: (<= (length$ ?s) 0))(OBJECT ?p-obj)(model ?m-obj))
    (MAIN::upcoming-household (person ?name))
    (MAIN::employment (employee ?name))
    (MAIN::household (id ?hh) (village ?village))
   	 =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule save-up-for-house for " ?name " of " ?hh )))
    ;; look for existing stokvels any friends belong to
	(bind ?s-list (create$))
    (bind ?qr (run-query* MAIN::stokvels-query ?name ?village))
	(while (?qr next)
		(bind ?s-list (create$ ?s-list (?qr get friend-st)))
    )
    (if (<= (length$ ?s-list) 0) then
        ;; try to form a new stokvel
        (assert 
        (MAIN::group-wanted
            (person ?name)
            (type stokvel)
            (tickStamp (current-tick))))
    else
        ;; join the stokvel where most of my friends are 
        ;; --> that's the stokvel featuring prominently in the list
        (bind ?my-stokvel (element-with-highest-occurrence ?s-list))
        (bind ?my-stokvel-obj nil)
        (bind ?qr (run-query* MAIN::specific-stokvel-query ?my-stokvel))
        (while (?qr next) (bind ?my-stokvel-obj (?qr get obj)))
        (dump (str-cat ?name " joins " ?my-stokvel))
        (call ?my-stokvel-obj addMember ?p-obj)
        (call ?m-obj recordStokvelEvent ?my-stokvel 
            (str-cat ?name " (" ?age ") chooses " ?my-stokvel " from " ?s-list " to join to save up for a house") ?p-obj)
    )
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule save-up-for-childrens-education
	"a female adult will try to form a stokvel if there are children >= 13 in the household"
    (MAIN::person  (alive TRUE)(name ?name)(gender female)(householdID ?hh)(atSchool FALSE){age > 28}
        (stokvels ?s &: (<= (length$ ?s) 0)) (OBJECT ?p-obj)(model ?m-obj))
    (MAIN::household (id ?hh) (village ?village)
        (spendingCommitments ?commitments)(expectedIncome ?income)(cash ?cash))
    (exists (MAIN::person (householdID ?hh){age >= 13}(atSchool TRUE)))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule save-up-for-childrens-education for " ?name " of " ?hh )))
    (bind ?reserve (- (+ ?income ?cash) ?commitments))
    (if (and (> ?reserve 30) (< ?reserve 300)) then
	    ;; look for existing stokvels any friends belong to
		(bind ?s-list (create$))
	    (bind ?qr (run-query* MAIN::stokvels-query ?name ?village))
		(while (?qr next)
			(bind ?s-list (create$ ?s-list (?qr get friend-st)))
	    )
	    (if (<= (length$ ?s-list) 0) then
	        ;; try to form a new stokvel
	        (assert 
	        (MAIN::group-wanted
	            (person ?name)
	            (type stokvel)
	            (tickStamp (current-tick))))
	    else
	        ;; join the stokvel where most of my friends are 
	        ;; --> that's the stokvel featuring prominently in the list
	        (bind ?my-stokvel (element-with-highest-occurrence ?s-list))
	        (bind ?my-stokvel-obj nil)
	        (bind ?qr (run-query* MAIN::specific-stokvel-query ?my-stokvel))
	        (while (?qr next) (bind ?my-stokvel-obj (?qr get obj)))
	        (dump (str-cat ?name " joins " ?my-stokvel))
	        (call ?my-stokvel-obj addMember ?p-obj)
	        (call ?m-obj recordStokvelEvent ?my-stokvel 
            	(str-cat ?name " (" ?age ") chooses " ?my-stokvel " from " ?s-list " to join to save up for children's education") ?p-obj)                       
	    )
    ) 
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
           
)

(defrule create-stokvel
	"create a new stokvel from a group of friends"
    (MAIN::model (OBJECT ?m-obj))
    ?f <- (MAIN::group (group-id ?leader) (group-type stokvel) (members $?members))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule create-stokvel for " ?leader )))
    ;; fetch Java objects of members
    (bind ?list (new java.util.ArrayList))
    (foreach ?member ?members
		(bind ?qr (run-query* MAIN::person-object-query ?member))
        (while (?qr next)
        	(?list add (?qr get object)))
     )  
    (dump (str-cat "create new stokvel for " ?members))  
    ;; create stokvel as new shadow fact
    (new Stokvel ?list ?m-obj) ;; will be put into Jess from Java (in constructor of Stokvel)
    (retract ?f)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
) 

(defrule payout-savings
    "payout accumulated savings to members of stokvel (starts a new cycle automatically)"
    (MAIN::model (week 3)(tick ?tick) (OBJECT ?m-obj))
    (MAIN::stokvel (payOutDate ?tick)(memberIDs $?members)(name ?name)(OBJECT ?s-obj))
    =>
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule payout-savings for " ?name )))
    (bind ?amounts (call ?s-obj payOut))
    (for (bind ?i 1) (<= ?i (length$ ?amounts)) (++ ?i)
        (assert
            (MAIN::payment
                (amount (nth$ ?i ?amounts))
                (cause stokvel-payout)
                (payment-date ?tick)
                (recipient (nth$ ?i ?members)))))
    (dump (str-cat ?name " pays out savings to members"))
    (call ?m-obj recordStokvelEvent ?name (str-cat ?name " pays out savings " ?amounts " to members " ?members " at tick " ?tick))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
    
)  

(defrule pay-monthly-fee-to-stokvel
	"every member needs to pay the monthly fee at the beginning of each month"
    (declare (no-loop TRUE))
    (MAIN::model (week 0)(OBJECT ?m-obj))
    (MAIN::person  (alive TRUE)(name ?member) (householdID ?hh) (OBJECT ?p-obj))
    (MAIN::stokvel (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) 
        (monthlyFee ?fee) (membersOnGrace $?members-on-grace &: (not (member$ ?member $?members-on-grace)))(OBJECT ?bs-obj))
    ?f <- (MAIN::household (id ?hh) (cash ?cash &: (> ?cash ?fee)) (spendingCommitments ?commitments))
    (not (MAIN::payment (payer ?member) (recipient ?bs) (cause group-fee)))
	=>
	(if (> (current-tick) 350) then
        ((System.out) println (str-cat "rule pay-monthly-fee-to-stokvel for " ?member " of " ?hh )))
    ;    ((System.out) println (str-cat ?member " pays monthly fee of " ?fee " Rand to " ?bs ))
    (dump (str-cat ?member " pays monthly fee of " ?fee " Rand to " ?bs ))
    (modify ?f (spendingCommitments (+ ?commitments ?fee)))
    (bind ?tick (current-tick))
    (assert
        (MAIN::payment
        	(payer ?member)
    		(recipient ?bs)
    		(cause group-fee)
            (amount ?fee)
    		(payment-date ?tick)))
	(call ?bs-obj payIn ?p-obj ?fee)
    (call ?m-obj recordStokvelEvent ?bs (str-cat ?member " pays monthly fee of " ?fee " at tick " ?tick))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule request-grace-period
	"if there's not enough cash ask for a grace period"
    (MAIN::model (week 0)(OBJECT ?m-obj))
    (MAIN::person (alive TRUE) (name ?member) (householdID ?hh) (OBJECT ?p-obj))
    (MAIN::stokvel (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) 
        (monthlyFee ?fee) (membersOnGrace $?members-on-grace &: (not (member$ ?member $?members-on-grace))) (OBJECT ?bs-obj))
    (MAIN::household (id ?hh) (cash ?cash &: (<= ?cash ?fee)))      
    (not (MAIN::payment (payer ?member) (recipient ?bs) (cause fee)))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule request-grace-period for " ?member " of " ?hh )))
    ;    ((System.out) println (str-cat ?member " requests grace period from " ?bs ))
    (dump (str-cat ?member " requests grace period from " ?bs ))
	(call ?bs-obj requestGracePeriod ?p-obj)
    (call ?m-obj recordStokvelEvent ?bs (str-cat ?member " requests grace period at tick " ?tick))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule end-grace-period
	"either pay off debt or leave"
    (declare (no-loop TRUE))
    (MAIN::model (week 0)(OBJECT ?m-obj))
    (MAIN::person (alive TRUE) (name ?member) (householdID ?hh) (OBJECT ?p-obj))
    (MAIN::stokvel (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) 
        (monthlyFee ?fee) (membersOnGrace $?members-on-grace &: (member$ ?member $?members-on-grace)) 
    	(gracePeriods $?grace-periods)(OBJECT ?bs-obj))
    ?f <- (MAIN::household (id ?hh) (cash ?cash) (spendingCommitments ?commitments))          
     =>
	(if (> (current-tick) 350) then
        ((System.out) println (str-cat "rule end-grace-period for " ?member " of " ?hh )))
    ;; find out how long member has been on grace period
    (bind ?tick (current-tick))
    (bind ?num-months (/ (- ?tick (nth$ (member$ ?member $?members-on-grace) $?grace-periods)) 4)) 
    (if (> ?num-months 0) then 
	    (if (<= ?num-months 5) then
	         ;; check if it's possible to pay off debt now (plus pay fee of this month)
            (bind ?amount (* ?fee (+ ?num-months 1)))  
		    (if (> ?cash ?amount) then
;			    ((System.out) println (str-cat ?member " pays off grace period debt of " ?amount " Rand to " ?bs ))
                (dump (str-cat ?member " pays off grace period debt of " ?amount " Rand to " ?bs ))
                (call ?m-obj recordStokvelEvent ?bs 
                    (str-cat ?member " pays off grace period debt of " ?amount " after " ?num-months " at tick " ?tick))
			    (call ?bs-obj payIn ?p-obj ?amount)
	            (modify ?f (spendingCommitments (+ ?commitments ?amount)))
			    (assert
			        (MAIN::payment
			        	(payer ?member)
			    		(recipient ?bs)
			            (cause fee)
			            (payment-date ?tick)
			            (amount ?amount)		
			    ))           	            
		    )        
	     else 
	        ;; leave stokvel
;		    ((System.out) println (str-cat ?member " leaves " ?bs " due to lack of money" ))
            (dump (str-cat ?member " leaves " ?bs " due to lack of money" ))
		    (call ?bs-obj removeMember ?p-obj)
            (call ?m-obj recordStokvelEvent ?bs (str-cat ?member " leaves due to lack of money at tick " ?tick))
	     )
	)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule disband-stokvel
    "stokvels disband when there are not enough members left"
    (MAIN::model (OBJECT ?m-obj))
    ?s <- (MAIN::stokvel (name ?stokvel)(memberIDs $?members &: (< (length$ ?members) 4))(OBJECT ?s-obj))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule disband-stokvel for " ?stokvel )))
    ;; distribute accumulated savings
    (bind ?amounts (call ?s-obj payOut))
    (for (bind ?i 1) (<= ?i (length$ ?amounts)) (++ ?i)
        (assert
            (MAIN::payment
                (amount (nth$ ?i ?amounts))
                (cause stokvel-payout)
                (payment-date ?tick)
                (recipient (nth$ ?i ?members)))))
    (call ?m-obj recordStokvelEvent ?stokvel 
        (str-cat "Stokvel disbands because there aren't enough members left." crlf "Pays out " ?amounts " to remaining members " ?members))
    ;; disband stokvel --> removes stokvel from each member's list of stokvels 
    (call ?s-obj disband)
    ;; retract shadow fact
    (retract ?s) 
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
           
)
 
;; --------------- sexual partner related rules -----------------

(defrule evaluate-partner-request
	"female accepts or rejects advances of males"
    (declare (no-loop TRUE))
    (MAIN::person (alive TRUE) (name ?female) (currentPartners $?partners) (maxNumPartners ?max &: (> ?max 0))
        (agentEndorsementScheme ?end-scheme) (tag $?tag) (OBJECT ?female-obj))
    (exists (MAIN::message (header sexual-partner-request) (recipient ?female) (sender ?male-person)))
	=>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule evaluate-partner-request for " ?female)))
    ;; find all requests
    (bind ?persuers (create$))
    (bind ?qr (run-query* MAIN::messages-query ?female sexual-partner-request ?tick))
    (while (?qr next)
        (bind ?persuer (?qr get sender))
        (if (not (member$ ?persuer ?partners)) then 	;; only consider new partners
        	(bind ?persuers (create$ ?persuers ?persuer)))) 
    (if (> (length$ ?persuers) 0) then
	    ;; evaluate persuers using endorsements
	    (bind ?end-values (create$))
	    (foreach ?male ?persuers
	        (bind ?qr (run-query* MAIN::endorsements-query ?female ?male ?tick))
	        (bind ?endorsements (create$))
	        (while (?qr next)
	         	(bind ?endorsements (create$ ?endorsements (?qr get endorsement)))
	        )
	        (bind ?end-values (create$ ?end-values (endorsement-value ?end-scheme ?endorsements)))
	    )
;;	    (dump (str-cat ?female " has " (length$ ?persuers) " persuers to choose from"))
	    ;; sort list of endorsement values in descending order of values
	    (bind ?indices (sort-values ?end-values TRUE))
	    ;; check the first persuer
	    (bind ?male (nth$ (nth$ 1 ?indices) ?persuers))
	    (bind ?end-value (nth$ (nth$ 1 ?indices) ?end-values))
	    (if (< (length$ ?partners) ?max) then 
	    	;; "space" available --> accept when value > threshold
		    ;; use tag[3] as threshold
		    (bind ?threshold (* (nth$ 3 ?tag) 5))    
		    (if (>= ?end-value ?threshold) then
		        (dump (str-cat ?female " accepts " ?male " as sexual partner and adds him to (" ?partners ")"))
		        (bind ?qr (run-query* MAIN::person-object-query ?male))
		        (while (?qr next) (bind ?male-obj (?qr get object)))
		        (call ?female-obj addCurrentPartner ?male-obj)
		        (call ?male-obj addCurrentPartner ?female-obj)
		     else
;;		        (dump (str-cat ?female " rejects advances of " ?male))   
		    )
	    else
	    	;; see if new male is better than one of the current partners
	        ;; evaluate current partners using endorsements (again) ### better store end-values in Person??
		    (bind ?end-values (create$))
		    (foreach ?partner ?partners
		        (bind ?qr (run-query* MAIN::endorsements-query ?female ?partner ?tick))
		        (bind ?endorsements (create$))
		        (while (?qr next)
		         	(bind ?endorsements (create$ ?endorsements (?qr get endorsement)))
		        )
		        (bind ?end-values (create$ ?end-values (endorsement-value ?end-scheme ?endorsements)))
	    	)
            ;; sort endorsement values in ascending order
            (bind ?indices (sort-values ?end-values FALSE))
            ;; compare new male to "weakest" partner
            (if (> ?end-value (nth$ (nth$ 1 ?indices) ?end-values)) then
                ;; accept new one, kick out old one
                (bind ?loser (nth$ (nth$ 1 ?indices) ?partners))
                (dump (str-cat ?female " accepts " ?male " as sexual partner and gives " ?loser " the pass out of (" ?partners ")"))
                (bind ?qr (run-query* MAIN::person-object-query ?male))
		        (while (?qr next) (bind ?male-obj (?qr get object)))
		        (call ?female-obj addCurrentPartner ?male-obj)
		        (call ?male-obj addCurrentPartner ?female-obj)
                (bind ?qr (run-query* MAIN::person-object-query ?loser))
		        (while (?qr next) (bind ?loser-obj (?qr get object)))
		        (call ?female-obj removeCurrentPartner ?loser-obj)
		        (call ?loser-obj removeCurrentPartner ?female-obj)
            else
                ;; reject new one
;;                (dump (str-cat ?female " rejects advances of " ?male " in favour of current partners"))    
            )                       
	))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
 )

;; ---------------- mine related rules --------------------------

(defrule advertise-mine-jobs
    "a mine advertises jobs when it needs workers"
    (MAIN::mine (name ?mine) (neededUnskilled ?numNeeded) (numUnskilled ?numWorkers &: (< ?numWorkers ?numNeeded)))
    (not (MAIN::employees-demanded (employer ?mine)))
    (MAIN::model (week 0))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule advertise-mine-jobs for " ?mine )))
    (watch facts)
    (assert
        (MAIN::employees-demanded
        	(employer ?mine)
    		(number-demanded (- ?numNeeded ?numWorkers))
    		(tickStamp ?tick)))
    (unwatch facts)
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(defrule look-for-shack
    "a miner looks for a shack to rent in the village"
    (MAIN::person (name ?miner){age > 29}(householdID nil))
    (not (MAIN::message (header shack-application)(sender ?miner)))	;; make sure they only apply to one at a time
    (not (MAIN::rent-a-shack (tenant ?miner)))	;; make sure they don't apply when they already rent a shack
    (MAIN::message (header shack-for-rent)(sender ?hh)(tickStamp ?tick)(slot-names $?slots)(slot-values $?values))
    (MAIN::employment (employee ?miner)(employer ?mine){end-date > ?tick})
    (MAIN::mine (name ?mine))
    =>
    (bind ?tick (current-tick))
	(if (> ?tick 350) then
        ((System.out) println (str-cat "rule look-for-shack for " ?miner)))
        
    
;    (dump (str-cat ?miner " applies to rent shack from " ?hh))
    ;; just ignore the rent for now
    (assert
        (MAIN::message
            (header shack-application)
            (recipient ?hh)
            (sender ?miner)
            (slot-names (create$ end-tick))
            (slot-values (create$ ?end-date))
            (tickStamp ?tick)))
	(if (> ?tick 350) then
        ((System.out) println "end of rule"))    
        
)

(set-current-module MAIN)