;; this file contains the definitions of queries
;; used in the rules to find a particular sub-set
;; of facts

(require deftemplates)
(provide defqueries)

;;---- endorsement and tag related ---------------------

(defquery endorsements-query
    (declare (variables ?endorser ?endorsee ?tick))
    (endorsed
            (endorsement ?endorsement)
            (endorsed-object ?endorsee)
            (uniquifiers $?uniqs)
            (clauseOwner ?endorser)
            (tickStamp ?t&: (<= ?t ?tick))))

(defquery specific-endorsement-query
    (declare (variables ?endorser ?endorsement ?tick))
    (endorsed
        (endorsement ?endorsement)
        (endorsed-object ?endorsee)
        (uniquifiers $?uniqs)
        (clauseOwner ?endorser)
        (tickStamp ?t&: (<= ?t ?tick)))
)

(defquery object-type-endorsement-value-query
	"collect endorsement values on agents"
	(declare (variables ?agent ?type ?tick))
    (object-endorsement-value
        (object ?object)
        (object-type ?type)
        (endorsement-value ?end-value)
        (clauseOwner ?agent)
        (tickStamp ?tick)))

(defquery similarity-index-query
	"similarity indices and corresponding persons for an agent"
	(declare (variables ?agent ?tick))
    (similarity-index
        (clauseOwner ?agent)
        (similarity ?similarity)
        (other-person ?other-person)
        (tickStamp ?tick))
)

;; ---- group related ----------------------------------

(defquery burial-societies-query
	"find the burial societies my friends and relatives from the village belong to"
	(declare (variables ?person ?village))
    (endorsed
        (endorsement is-friend | is-kin)
        (clauseOwner ?person)
        (endorsed-object ?friend))
    (person (name ?friend) (householdID ?hh)(burialSocieties $?friend-bs &: (> (length$ ?friend-bs) 0)))
    (household (id ?hh) (village ?village))
)

(defquery specific-burial-society-query
    "find a particular burial society"
    (declare (variables ?name))
    (burialSociety (name ?name) (monthlyFee ?fee) (memberIDs $?members)(OBJECT ?obj))
)

(defquery stokvels-query
	"find the stokvels my friends from the village belong to"
	(declare (variables ?person ?village))
    (endorsed
        (endorsement is-friend)
        (clauseOwner ?person)
        (endorsed-object ?friend))
    (person (name ?friend) (householdID ?hh)(stokvels $?friend-st &: (> (length$ ?friend-st) 0)))
    (household (id ?hh) (village ?village))
)

(defquery specific-stokvel-query
    "find a particular stokvel"
    (declare (variables ?name))
    (stokvel (name ?name) (monthlyFee ?fee) (memberIDs $?members)(OBJECT ?obj))
)


;; ---- marriage related -------------------------------

(defquery suitable-partner-query
    "find unmarried females of the right age range that are NOT of the same household"
    (declare (variables ?hh ?male-age))
    (person  (alive TRUE)(migrant FALSE)(gender female)(householdID ~?hh)(maritalStatus ~married)(name ?name) 
        (atSchool FALSE)
        (age ?age &: (and (> ?age 15) (<= ?age ?male-age))))
)

;; ---- migration related ------------------------------

(defquery migrant-relatives-query
    "find relatives who are migrant"
    (declare (variables ?person ?hh))
    (or (person (name ?relative) (migrant TRUE))
        (person (name ?relative) (onVisit TRUE)))
    (or (endorsed (clauseOwner ?person) (endorsed-object ?relative) (endorsement is-kin))
        (household (id ?hh) (memberIDs $? ?relative $?))
    )
)

(defquery migrant-friends-query
    "find friends who are migrant"
    (declare (variables ?person))
    (endorsed (clauseOwner ?person) (endorsed-object ?friend) (endorsement is-friend))
    (or (person (name ?friend) (migrant TRUE))
        (person (name ?friend) (onVisit TRUE)))
)

;; ---- person related ---------------------------------

(defquery person-object-query
	"to get the Java object from the name of the agent. useful in building lists on rule RHS"
	(declare (variables ?agent-name))
    (person (name ?agent-name)(OBJECT ?object)))

(defquery person-household-query
    "find a person's household"
    (declare (variables ?person))
    (person (name ?person) (householdID ?hh))
)


;; ---- communication related --------------------------

(defquery messages-query
	"collect all messages for recipient with specified header"
	(declare (variables ?recipient ?header ?tick))
	(message
        (recipient ?recipient)
        (tickStamp ?tick)
        (header ?header)
        (sender ?sender)
        (slot-names $?slot-names)
        (slot-values $?slot-values)))

;; ---- employment related -----------------------------

(defquery employment-query
	"find all employments of an agent"
	(declare (variables ?agent))
    (employment
        (employee ?agent)
        (employer ?employer)
        (wage ?wage))
	)

(defquery job-application-query
    "find all applications for my advertised jobs"
    (declare (variables ?employer ?tick))
    (job-application
        (applicant ?applicant)
        (potential-employer ?employer)
        (tickStamp ?tick))
)

(defquery job-performance-query
    "find previous assessments of applicants"
    (declare (variables ?employer ?employees))
    (job-performance
        (employer ?employer)
        (employee ?employee &: (member$ ?employee ?employees))
        (reliability ?r)
        (capability ?c)
        (num-jobs ?n)
        (tickStamp ?t))
)



