(require deftemplates)
(require defqueries)

(set-current-module INIT)

;(defrule print-module-init
;    (declare (salience 100))
;    (MAIN::model (tick ?tick))
;    =>
;    (dump (str-cat "MODULE INIT started")))


;; -------- endorsements related rules -------------

(defrule similarity-identification
	"identify others with most similar and similar tags"
    (MAIN::person (alive TRUE)(tag $?own-tag)(name ?person)(gender ?gender)(knownPersons $?known-persons)(age ?own-age))
	(MAIN::person
        (alive TRUE)
        (name ?other &: (member$ ?other ?known-persons))
        (gender ?gender)
        (age ?other-age &: (< (abs (- ?own-age ?other-age)) (* ?own-age ?*similarity-age-range*))) ;; age range is +- 25% of own age
        (tag $?other-tag))
    (not
        (MAIN::similarity-index (clauseOwner ?person)(other-person ?other)))
    =>
    (bind ?similarity (number-of-common-attributes ?own-tag ?other-tag))
;    (watch facts)
    (assert
        (similarity-index 
            (other-person ?other)
            (similarity ?similarity)
            (clauseOwner ?person)
            (tickStamp (current-tick))))
;    (unwatch facts)
)

(defrule endorse-as-trustworthy
	"reliability + honesty = trustworthiness"
;    (model (OBJECT ?m-obj))
    (endorsed (clauseOwner ?endorser)(endorsed-object ?endorsee)(endorsement reliable))
    (endorsed (clauseOwner ?endorser)(endorsed-object ?endorsee)(endorsement honest))
    (not (endorsed (clauseOwner ?endorser)(endorsed-object ?endorsee)(endorsement trustworthy)))
	=>
    (assert
        (endorsed
            (endorsed-object ?endorsee)
            (endorsement trustworthy)
            (clauseOwner ?endorser)
            (tickStamp (current-tick))))
)

(defrule endorse-same-church
    "agents from different household and same age group"
;	(MAIN::model (OBJECT ?m-obj))
    (MAIN::person (householdID ?hh1)(age ?age1)(name ?person1))
    (MAIN::person (householdID ?hh2 &:(neq ?hh2 ?hh1))
        (age ?age2 &: 
            (and
                (<= (abs (- ?age1 ?age2)) 5)
                (<= ?age1 ?age2))) ;;so only need fire once unless two ages the same
        (name ?person2))
    (MAIN::household (id ?hh1)(church ?church))
    (MAIN::household (church ?church)(id ?hh2 &:(neq ?hh2 ?hh1)))
    (not (endorsed (endorsement same-church) (clauseOwner ?person1|?person2)(endorsed-object ?person2|?person1)))
    =>
;    (watch facts)
    (bind ?tick (current-tick))
    (assert
        (endorsed
            (endorsement same-church)
            (endorsed-object ?person2)
            (clauseOwner ?person1)
            (tickStamp ?tick))
        (endorsed
            (endorsement same-church)
            (endorsed-object ?person1)
            (clauseOwner ?person2)
            (tickStamp ?tick)))
;    (unwatch facts)
)

(defrule endorse-same-denomination
    "endorse other agents already known since sharing denomination
    is not a reason for acquaintanceship"
;    (MAIN::model (OBJECT ?m-obj))
;    (MAIN::known-person (known ?acquaintance)(clauseOwner ?person)(tickStamp 0))
    (MAIN::person (householdID ?hh1)(name ?person)(knownPersons $?known-persons))
    (MAIN::person (name ?acquaintance &: (member$ ?acquaintance ?known-persons))(householdID ?hh2 &: (neq ?hh2 ?hh1)))
    (MAIN::household (id ?hh1)(church ?church1))
    (MAIN::household (id ?hh2)(church ?church2))
    (MAIN::church (name ?church1)(denomination ?denom))
    (MAIN::church (name ?church2)(denomination ?denom))
    (not (endorsed (endorsement same-denomination)(clauseOwner ?person)(endorsed-object ?acquaintance)))
	=>
    (assert
        (endorsed
            (endorsement same-denomination)
            (endorsed-object ?acquaintance)
            (clauseOwner ?person)
            (tickStamp (current-tick))))
)



;; -------- household related rules ----------------

(defrule food-payment
	"buy monthly bulk food"
    (declare (no-loop TRUE))
	(MAIN::model (week 0)(OBJECT ?m-obj))
    ?f <- (MAIN::household (id ?hh) (cash ?cash) (monthlyFoodCost ?food-cost) (spendingCommitments ?commitments))      
    (not (MAIN::payment (payer ?hh) (cause food))) ;; this is needed to prevent loops when other rules change the commitments  
     =>
    (bind ?tick (current-tick))
    (if (< ?cash ?food-cost) then
    	(watch facts)
        ;; will have to ask neighbours and friends for help
        (assert
            (MAIN::help-needed
            	(household ?hh)
        		(amount (- ?food-cost ?cash))
        		(tickStamp ?tick)))
        (assert
            (MAIN::asked-for-help
                (requester ?hh)
                (hh-list (create$))
                (tickStamp ?tick)))
        (unwatch facts)
        (call ?m-obj incNumHHAskingForFood)
    )
    (bind ?amount (min ?food-cost ?cash))
    (assert
        (MAIN::payment
            (payer ?hh)
            (cause food)
            (amount ?amount)
            (payment-date ?tick)))
    (modify ?f (spendingCommitments (+ ?commitments ?amount)))
;    ((System.out) println (str-cat ?hh " buys bulk food for " ?amount))
    ;;*;(dump (str-cat ?hh " buys bulk food for " ?amount))
)

(defrule burial-payment
    "if a member of the household dies the household has to arrange a funeral"
    (declare (no-loop TRUE))
;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::death-in-hh (household ?hh) (deceased ?name)(tickStamp ?tick))
;;    (MAIN::death (household ?hh) (tickStamp =(- ?tick 1)))	;; assume the funeral is one tick later than the death
    ?f <- (MAIN::household (id ?hh) (cash ?cash) (spendingCommitments ?commitments))
    (not (MAIN::payment (payer ?hh)(cause burial){payment-date == (+ ?tick 1)}))	;; needed to prevent loops when other rules change the commitments
	=>
    (bind ?amount (min ?*burial-cost* ?cash)) ;; find out real burial costs ###
            ;; funerals can sometimes cost the equivalent of seven months' income. Source: http://knowledge.allianz.com/en/globalissues/microfinance/microcredit/financial_diaries_microfinance.html
            ;; 10000-15000 Rand. Source: http://www.financialdiaries.com/files/Funerals%20and%20Finance%20-%20Events%20in%20the%20lives%20of%20Financial%20Diaries%20households.pdf 
    		;; from last field trip to Ga-Selala: 4000-10000 Rand
    (modify ?f (spendingCommitments (+ ?commitments ?amount)))
    (assert
        (MAIN::payment
        	(payer ?hh)
    		(cause burial)
    		(amount ?amount)
          	(payment-date (current-tick))))
;	((System.out) println (str-cat ?hh " pays " ?amount " for burial of " ?name))
    (dump (str-cat ?hh " pays " ?amount " for burial of " ?name))
)

(defrule burial-contribution-to-kin
    "contribute to burial for related household (only once per household)"
    (declare (no-loop TRUE))
;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::death (household ?hh) (name ?name)(tickStamp ?tick))
    (MAIN::endorsed (endorsement is-kin) (endorsed-object ?name) (clauseOwner ?relative))
    (MAIN::household (headID ?relative) (id ?contributing-hh) (spendingCommitments ?commitments)(OBJECT ?contributing-hh-obj))
    (not (MAIN::payment (payer ?contributing-hh) (recipient ?hh)(cause burial)(payment-date ?tick))) ;; needed to prevent loops when other rules change the spendingCommitments
     =>
    (bind ?amount (call ?contributing-hh-obj contributeToBurial ?name ?hh TRUE))
    (call ?contributing-hh-obj setSpendingCommitments (+ ?commitments ?amount))
    (assert
        (MAIN::payment
         	(payer ?contributing-hh)
            (recipient ?hh)
            (cause burial)
            (payment-date ?tick)
            (amount ?amount)
        )
    )
;	((System.out) println (str-cat ?contributing-hh " contributes " ?amount " to burial costs of related " ?hh))
    (dump (str-cat ?contributing-hh " contributes " ?amount " to burial costs of related " ?hh))
)

(defrule burial-contribution-to-neighbour
    "contribute to burial for household in same village but not related (only once per household)"
    (declare (no-loop TRUE))
;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::death (household ?hh) (name ?name)(tickStamp ?tick))
    (MAIN::household (id ?hh) (village ?village))
    (MAIN::household (id ?contributing-hh) (village ?village) (headID ?contributing-head) (spendingCommitments ?commitments)(OBJECT ?contributing-hh-obj))
    (not (MAIN::endorsed (endorsement is-kin) (endorsed-object ?name) (clauseOwner ?contributing-head)))
    (not (MAIN::payment (payer ?contributing-hh) (recipient ?hh)(cause burial)(payment-date ?tick))) ;; needed to prevent loops when other rules change the spendingCommitments
	=>
    (bind ?amount (call ?contributing-hh-obj contributeToBurial ?name ?hh FALSE))
    (call ?contributing-hh-obj setSpendingCommitments (+ ?commitments ?amount))
    (assert
        (MAIN::payment
         	(payer ?contributing-hh)
            (recipient ?hh)
            (cause burial)
            (payment-date ?tick)
            (amount ?amount)
        )
    )
;	((System.out) println (str-cat ?contributing-hh " contributes " ?amount " to burial costs of neighbour " ?hh))
    (dump (str-cat ?contributing-hh " contributes " ?amount " to burial costs of neighbour " ?hh))
)

(defrule buy-additional-food
    "buy food from the borrowed money"
    (declare (no-loop TRUE))
    (MAIN::model (week 1|2|3))
    ?f <- (MAIN::help-needed (amount ?amount) (household ?hh) (tickStamp ?tick))
    ?h <- (MAIN::household (id ?hh) (cash ?cash) (spendingCommitments ?commitments) (model ?m-obj))
    =>
    (bind ?additional-food (max (min ?amount ?cash) 0))
    (assert
        (MAIN::payment
        	(amount ?additional-food)
    		(cause food)
    		(payer ?hh)
    		(payment-date (+ ?tick 1))))
	(modify ?h (spendingCommitments (+ ?commitments ?additional-food)))
;    (retract ?f)
    (dump (str-cat ?hh " buys additional food for " ?additional-food " Rand"))
    (if (< ?additional-food ?amount) then
        (modify ?f (amount (- ?amount ?additional-food))(tickStamp (+ ?tick 1)))
        (call ?m-obj incNumHHWithoutEnoughFood)
     else
        (retract ?f)
     )
)


;; -------- burial society related rules -----------

(defrule payout-for-death-of-member
    "burial society pays out to household of deceased who was a member of the society"
    (declare (no-loop true))
;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::death (household ?hh) (name ?deceased) (tickStamp ?tick))
    (MAIN::burialSociety (name ?bs)(memberIDs $?member-list &: (member$ ?deceased $?member-list)) (OBJECT ?bs-obj))
    (not (MAIN::payment (payer ?bs) (recipient ?hh) (cause burial-payout) (payment-date ?tick))) ;; needed to prevent loops when other rules change the shadow fact (burialSociety)
	=>
    (bind ?amount (call ?bs-obj payOut ?deceased ?deceased))
    (assert
        (MAIN::payment
        	(payer ?bs)
    		(recipient ?hh)
            (cause burial-payout)
            (payment-date ?tick)
            (amount ?amount)
    ))
;	((System.out) println (str-cat ?bs " pays out " ?amount " Rand to " ?hh " for the death of its member " ?deceased))
    (dump (str-cat ?bs " pays out " ?amount " Rand to " ?hh " for the death of its member " ?deceased))
)

(defrule payout-for-covered-death
	"burial society pays out to household of deceased if a household member is member of the society"
    (declare (no-loop true))
    ;    (MAIN::model (OBJECT ?m-obj))
    (MAIN::death (household ?hh) (name ?deceased) (tickStamp ?tick))
    (MAIN::person (householdID ?hh) (name ?member &: (neq ?deceased ?member)))
    (MAIN::burialSociety (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) (OBJECT ?bs-obj))
	(not (MAIN::payment (payer ?bs)(recipient ?member)(cause burial-payout)(payment-date ?tick))) ;; needed to prevent loops
    =>
    (bind ?amount (call ?bs-obj payOut ?member ?deceased))
    (assert
        (MAIN::payment
        	(payer ?bs)
    		(recipient ?member)
            (cause burial-payout)
            (payment-date ?tick)
            (amount ?amount)
    ))
;	((System.out) println (str-cat ?bs " pays out " ?amount " Rand to " ?member " for the death of " ?deceased ))
    ;;*;(dump (str-cat ?bs " pays out " ?amount " Rand to " ?member " for the death of " ?deceased ))
)

(defrule pay-monthly-fee-to-burial-society
	"every member needs to pay the monthly fee at the beginning of each month"
    (declare (no-loop TRUE))
    (MAIN::model (week 0))
    (MAIN::person (name ?member) (householdID ?hh) (OBJECT ?p-obj))
    (MAIN::burialSociety (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) 
        (monthlyFee ?fee) (membersOnGrace $?members-on-grace &: (not (member$ ?member $?members-on-grace)))(OBJECT ?bs-obj))
    ?f <- (MAIN::household (id ?hh) (cash ?cash &: (> ?cash ?fee)) (spendingCommitments ?commitments))
    (not (MAIN::payment (payer ?member) (recipient ?bs) (cause group-fee)))
	=>
;    ((System.out) println (str-cat ?member " pays monthly fee of " ?fee " Rand to " ?bs ))
    ;;*;(dump (str-cat ?member " pays monthly fee of " ?fee " Rand to " ?bs ))
    (modify ?f (spendingCommitments (+ ?commitments ?fee)))
    (assert
        (MAIN::payment
        	(payer ?member)
    		(recipient ?bs)
    		(cause group-fee)
            (amount ?fee)
    		(payment-date (current-tick))))
	(call ?bs-obj payIn ?p-obj ?fee)
)

(defrule request-grace-period
	"if there's not enough cash ask for a grace period"
    (MAIN::model (week 0))
    (MAIN::person (name ?member) (householdID ?hh) (OBJECT ?p-obj))
    (MAIN::burialSociety (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) 
        (monthlyFee ?fee) (membersOnGrace $?members-on-grace &: (not (member$ ?member $?members-on-grace))) (OBJECT ?bs-obj))
    (MAIN::household (id ?hh) (cash ?cash &: (<= ?cash ?fee)))      
    (not (MAIN::payment (payer ?member) (recipient ?bs) (cause fee)))
    =>
;    ((System.out) println (str-cat ?member " requests grace period from " ?bs ))
    ;;*;(dump (str-cat ?member " requests grace period from " ?bs ))
	(call ?bs-obj requestGracePeriod ?p-obj)
)

(defrule end-grace-period
	"either pay off debt or leave"
    (declare (no-loop TRUE))
    (MAIN::model (week 0))
    (MAIN::person (name ?member) (householdID ?hh) (OBJECT ?p-obj))
    (MAIN::burialSociety (name ?bs)(memberIDs $?member-list &: (member$ ?member $?member-list)) 
        (monthlyFee ?fee) (membersOnGrace $?members-on-grace &: (member$ ?member $?members-on-grace)) 
    	(gracePeriods $?grace-periods)(OBJECT ?bs-obj))
    ?f <- (MAIN::household (id ?hh) (cash ?cash) (spendingCommitments ?commitments))          
     =>
	;; find out how long member has been on grace period
    (bind ?tick (current-tick))
    (bind ?num-months (/ (- ?tick (nth$ (member$ ?member $?members-on-grace) $?grace-periods)) 4)) 
    (if (> ?num-months 0) then 
	    (if (<= ?num-months 5) then
	         ;; check if it's possible to pay off debt now (plus pay fee of this month)
            (bind ?amount (* ?fee (+ ?num-months 1)))  
		    (if (> ?cash ?amount) then
;			    ((System.out) println (str-cat ?member " pays off grace period debt of " ?amount " Rand to " ?bs ))
                ;;*;(dump (str-cat ?member " pays off grace period debt of " ?amount " Rand to " ?bs ))
			    (call ?bs-obj payIn ?p-obj ?amount)
	            (modify ?f (spendingCommitments (+ ?commitments ?amount)))
			    (assert
			        (MAIN::payment
			        	(payer ?member)
			    		(recipient ?bs)
			            (cause fee)
			            (payment-date ?tick)
			            (amount ?amount)		
			    ))           	            
		    )        
	     else 
	        ;; leave burial society
;		    ((System.out) println (str-cat ?member " leaves " ?bs " due to lack of money" ))
            ;;*;(dump (str-cat ?member " leaves " ?bs " due to lack of money" ))
		    (call ?bs-obj removeMember ?p-obj)
	     )
	)
)

(defrule continue-burial-society-after-death
	"(new) household head decides to continue or not"
;    (MAIN::model (OBJECT ?m-obj))
    ?f <- (MAIN::membership-reminder (deceased ?deceased) (group ?bs) (type burial-society) (household ?hh))
	(MAIN::household (id ?hh) (cash ?cash) (spendingCommitments ?commitments))
    (MAIN::burialSociety (name ?bs) (monthlyFee ?fee) (OBJECT ?bs-obj))
    (MAIN::person (name ?name) (head TRUE) (householdID ?hh) (burialSocieties $?bs-list &: (not (member$ ?bs ?bs-list)))(OBJECT ?p-obj))
    =>  
    ;; continue membership if enough money
    (if (> (- ?cash ?commitments) ?fee) then
    	;; household head joins burial society
;        ((System.out) println (str-cat ?name " joins " ?bs " to continue membership of deceased " ?deceased " for " ?hh ))
        ;;*;(dump (str-cat ?name " joins " ?bs " to continue membership of deceased " ?deceased " for " ?hh ))
        (call ?bs-obj addMember ?p-obj)
        ;; get rid of reminder
        (retract ?f)
    )      
)               


;; -------- group related rules -------------------

(defrule create-draft-group
    "see how close proposed membership is for all proposers"
;    (MAIN::model (tick ?tick))
    (MAIN::group-wanted (person ?person) (type ?type))
    (not (MAIN::group (group-type ?type) (members $? ?person $?)))
    ?proposers <- (accumulate
        (bind ?list (create$))
        (bind ?list (create$ ?list ?proposer))
        ?list
        (MAIN::group-proposal
            (type ?type)
            (proposed-to ?person)
            (proposed-by ?proposer)))
    =>
    (if (eq ?type stokvel) then 
        ;;*;(dump (str-cat ?type" group proposed to "?person" by "?proposers))
    )
    (if (>= (length$ ?proposers) 4) then
    	;; start a group
;        (watch facts)
        (assert
            (MAIN::group 
            	(group-type ?type)
        		(members (create$ ?proposers ?person))
                (group-id ?person)
        ))
;		(unwatch facts)
    )
)

;; --------- employment related rules ------------------------

(defrule pick-employees
    "pick best from applicants to job offers"
    (declare (no-loop TRUE))
;	(MAIN::model (OBJECT ?m-obj) (standardWage ?w))
	?ed <-(MAIN::employees-demanded (employer ?employer) (number-demanded ?demand))
    (exists (MAIN::job-application (potential-employer ?employer)))
    ?f <- (MAIN::household (id ?employer) (cash ?cash)(spendingCommitments ?commitments &: (> (- ?cash ?commitments) (* ?demand 200))) (headID ?employer-head))
    (MAIN::person (name ?employer-head) (agentEndorsementScheme ?end-scheme))
;    (not (MAIN::employment (employer ?employer) (start-date ?tick))) ;; needed to prevent loops if other rules change the commitments
	=>
    (bind ?tick (current-tick))
	(bind ?qr (run-query* MAIN::job-application-query ?employer (- ?tick 1)))
	(bind ?applicants (create$))
	(while (?qr next)
        ;; only consider villagers, not miners
        (bind ?applicant (?qr get applicant))
        (if (str-index "person" ?applicant) then	
    		(bind ?applicants (create$ ?applicants ?applicant)))
    )
	(if (> (length$ ?applicants) ?demand) then
    	;; pick the best applicants, i.e. remove others from the list
    	;; use endorsements to evaluate each applicant
        (bind ?end-values (create$))
        (foreach ?applicant ?applicants
        	(bind ?qr (run-query* MAIN::endorsements-query ?employer-head ?applicant ?tick))
            (bind ?endorsements (create$))
            (while (?qr next)
            	(bind ?endorsements (create$ ?endorsements (?qr get endorsement)))
            )
        	(bind ?end-values (create$ ?end-values (endorsement-value ?end-scheme ?endorsements)))
        )
        ;; sort list of endorsement values in descending order of values
        (bind ?indices (sort-values ?end-values TRUE))
        ;; pick the first 1 - ?demand applicants
        (bind ?temp (create$))
        (bind ?i 1)
        (while (<= ?i ?demand)
        	(bind ?temp (create$ ?temp (nth$ (nth$ ?i ?indices) ?applicants)))
            (bind ?i (+ ?i 1))
        )
    	(bind ?applicants ?temp)
	)
	;; give remaining applicants a job
	(foreach ?applicant ?applicants
;    	((System.out) println (str-cat ?employer " employs " ?applicant " for a piece job"))
        (dump (str-cat ?employer " employs " ?applicant " for a piece job"))
        (bind ?end (+ ?tick 4)) ;; make more random
        (assert
    		(employment 
    			(employee ?applicant)
				(employer ?employer)
				(start-date ?tick)
				(end-date ?end)  
				(wage ?*standard-wage*) ;; adjust to length of employment
		))
		(bind ?qr (run-query* MAIN::person-object-query ?applicant))
        (while (?qr next)
        	(call (?qr get object) setLastJob ?end)
        )    
    )
    (modify ?f (spendingCommitments (+ ?commitments (* ?*standard-wage* (length$ ?applicants)))))
    (retract ?ed) ;; necessary to prevent loops if other rules change the commitments
)

;; ----------- sexual partner related rules --------------------------

(defrule find-sexual-partner-among-siblings-of-friends
    "look for potential partners among siblings of friends"
    (declare (no-loop TRUE))
    (MAIN::person (alive TRUE)(gender male)(name ?name)(migrant FALSE)(householdID ?hh)
        (age ?m-age &: (and (> ?m-age 14) (< ?m-age 65))) (maxNumPartners ?max)
        (currentPartners $?partners &: (<= (length$ ?partners) ?max))
    	(agentEndorsementScheme ?end-scheme) (tag $?tag))
    (MAIN::endorsed (clauseOwner ?name) (endorsed-object ?friend) (endorsement is-friend))
    (MAIN::person (alive TRUE)(name ?friend) (siblings $?siblings))
    (MAIN::person (alive TRUE)(gender female)(householdID ~?hh)(migrant FALSE)
        (name ?partner &: (and (member$ ?partner ?siblings) (not (member$ ?partner ?partners))))
        (maritalStatus ~married) {age > 14 && age <= ?m-age})
	=>
    (bind ?tick (current-tick))
    ;; evaluate partner's endorsement value
    (bind ?qr (run-query* MAIN::endorsements-query ?name ?partner ?tick))
    (bind ?endorsements (create$))
    (while (?qr next)
     	(bind ?endorsements (create$ ?endorsements (?qr get endorsement)))
    )
    (bind ?end-value (endorsement-value ?end-scheme ?endorsements))
    ;; if value > threshold then send "invitation" to partner
    ;; use tag[3] as threshold
    (bind ?threshold (* (nth$ 3 ?tag) 5))
    (if (>= ?end-value ?threshold) then
        (assert
        	(MAIN::message
        		(header sexual-partner-request)
    			(recipient ?partner)
    			(sender ?name)
        		(tickStamp ?tick)))
;;		(dump (str-cat ?name " (" ?m-age ") makes a pass at " ?partner "(" ?age ") -- endvalue = " ?end-value " threshold = " ?threshold))
	)  
)

(defrule find-sexual-partner-among-mutual-group
    "look for potential partners in same church/club/work and neighbours"
    (declare (no-loop TRUE))
    (MAIN::person (alive TRUE) (gender male) (name ?name) (householdID ?hh) (migrant FALSE)
        (age ?m-age &: (and (> ?m-age 14) (< ?m-age 65))) (maxNumPartners ?max)
    	(currentPartners $?partners &: (<= (length$ ?partners) ?max)) 
    	(agentEndorsementScheme ?end-scheme) (tag $?tag))
    (MAIN::endorsed (clauseOwner ?name) (endorsed-object ?partner) (endorsement same-church|same-club|same-work|is-neighbour))
    (MAIN::person (alive TRUE) (gender female) (migrant FALSE)
        (name ?partner &: (not (member$ ?partner ?partners))) 
        (householdID ~?hh) (maritalStatus ~married) {age > 14 && age <= ?m-age})
    =>
    (bind ?tick (current-tick))
    ;; evaluate partner's endorsement value
    (bind ?qr (run-query* MAIN::endorsements-query ?name ?partner ?tick))
    (bind ?endorsements (create$))
    (while (?qr next)
     	(bind ?endorsements (create$ ?endorsements (?qr get endorsement)))
    )
    (bind ?end-value (endorsement-value ?end-scheme ?endorsements))
    ;; if value > threshold then send "invitation" to partner
    ;; use tag[3] as threshold
    (bind ?threshold (* (nth$ 3 ?tag) 5))
    (if (>= ?end-value ?threshold) then
        (assert
        	(MAIN::message
        		(header sexual-partner-request)
    			(recipient ?partner)
    			(sender ?name)
        		(tickStamp ?tick)))
;;		(dump (str-cat ?name " (" ?m-age ") makes a pass at " ?partner "(" ?age ") -- endvalue = " ?end-value " threshold = " ?threshold))
	)  
)


(defrule find-sexual-partner-randomly
	"meet a woman randomly"
    (MAIN::person (alive TRUE) (gender male) (name ?name) (householdID ?hh) (migrant FALSE)
        (age ?m-age &: (and (> ?m-age 14) (< ?m-age 65))) (maxNumPartners ?max)
    	(currentPartners $?partners &: (<= (length$ ?partners) ?max)) (model ?m-obj)) 
    (test (probability-p ?*random-partner-probability*))    
	=>
    ;; pick a random partner
    (bind ?tick (current-tick))
	(bind ?list (create$))
	(bind ?qr (run-query* MAIN::suitable-partner-query ?hh ?m-age))
	(while (?qr next)
    	(bind ?list (create$ ?list (?qr get name)))
    )
	(if (> (length$ ?list) 0) then
    	(bind ?index (call ?m-obj randomInt 1 (length$ ?list)))
        (bind ?partner (nth$ ?index ?list))
        (assert
        	(MAIN::message
        		(header sexual-partner-request)
    			(recipient ?partner)
    			(sender ?name)
        		(tickStamp ?tick)))
		(dump (str-cat ?name " (" ?m-age ") makes a pass at randomly chosen " ?partner))
    )
)

(defrule show-off
    "a man with a job tells his potential female partner about this to impress her"
    (MAIN::message (header sexual-partner-request) (sender ?male) (recipient ?female) (tickStamp ?tick))
    (MAIN::employment (employee ?male) {end-date >= ?tick})
    (not (MAIN::endorsed (clauseOwner ?female)(endorsed-object ?male)(endorsement has-job)))
	=>
    ;; assert has-job endorsement for female
    (watch facts)
    (assert
        (MAIN::endorsed (clauseOwner ?female) (endorsed-object ?male) (endorsement has-job) (tickStamp ?tick))
  	)
	(unwatch facts)
)

(defrule break-up
    "every now and then end a relationship with a current partner"
    (declare (no-loop TRUE))
    (MAIN::person (name ?name)(maxNumPartners ?max)(currentPartners $?partners &: (> (length$ ?partners) 0)) 
        (age ?age)(model ?m-obj) (OBJECT ?p-obj))
    =>
    ;; since we don't know anything, just use a probability 
    ;; -- higher if number of partners is close to max
    ;; -- higher if younger?
    (bind ?break-up-prob (* 0.6 (/ (length$ ?partners) ?max) (/ (- 100 ?age) 100)))
    (if (call ?m-obj probabilityp ?break-up-prob) then
        ;; pick a random partner to remove
        (bind ?n (call ?m-obj randomInt 1 (length$ ?partners)))
        (dump (str-cat "rule break-up for " ?name " chose " ?n " from (" ?partners ") to break up with (prob. = " ?break-up-prob ")"))
        (if (<= ?n (length$ ?partners)) then
	        (bind ?loser (nth$ ?n ?partners))
	        (dump (str-cat ?name " breaks up with " ?loser))
            (bind ?qr (run-query* MAIN::person-object-query ?loser))
	        (while (?qr next) (bind ?loser-obj (?qr get object)))
	        (call ?p-obj removeCurrentPartner ?loser-obj)
	        (call ?loser-obj removeCurrentPartner ?p-obj)
	    )
    )
)


;; ----------- mining related ----------------------------

(defrule pick-mine-employees
    "pick best from applicants to job offers of a mine"
    (declare (no-loop TRUE))
	?ed <-(MAIN::employees-demanded (employer ?mine) (number-demanded ?demand))
    (MAIN::mine (name ?mine)(OBJECT ?mine-obj))
    (exists (MAIN::job-application (potential-employer ?mine)))
	=>
    ((System.out) println "rule pick-mine-employees fired")
    (bind ?tick (current-tick))
	(bind ?qr (run-query* MAIN::job-application-query ?mine (- ?tick 1)))
	(bind ?applicants (create$))
	(while (?qr next)
    	(bind ?applicants (create$ ?applicants (?qr get applicant)))
    )
	;; prefer men aged 23-36
    ;; allow for 5% women -- not yet implemented
    (call ?mine-obj pickUnskilledWorkers ?applicants ?tick)
	(retract ?ed)        
)

(defrule sack-worker-with-AIDS
    "remove worker when he has developed AIDS"
    ?e <- (MAIN::employment (employer ?mine)(employee ?worker))
    (MAIN::person (name ?worker)(hasAIDS TRUE))
    (MAIN::mine (name ?mine)(OBJECT ?mine-obj))
    =>
    (dump (str-cat ?mine " sacks worker " ?worker " because he has developed AIDS."))
    (call ?mine-obj sackWorker ?worker (current-tick))
    (retract ?e)
)

(defrule sack-worker-at-end-of-contract
    "remove worker when contract ended"
    ?e <- (MAIN::employment (employer ?mine)(employee ?worker)(end-date ?tick))
    (MAIN::mine (name ?mine)(OBJECT ?mine-obj))
    (MAIN::model (tick ?tick))
    =>
    (dump (str-cat ?mine " sacks worker " ?worker " because his contract ended."))
    (call ?mine-obj sackWorker ?worker ?tick)
    (retract ?e)
)

(defrule advertise-shack
	"a household advertises a shack to rent when they need money & have a free shack"
    (MAIN::household (id ?hh)(expectedIncome ?income){monthlyFoodCost > ?income}(model ?model-obj))
    (not (MAIN::message (header shack-for-rent)(sender ?hh)))
    (not (MAIN::rent-a-shack (landlord ?hh)))
    (test (?model-obj probabilityp ?*shack-proportion*)) ;; 40% of households have a shack
	=>
    ;; advertise the shack
    (assert
        (MAIN::message
            (header shack-for-rent)
            (sender ?hh)
            (slot-names (create$ rent))
            (slot-values (create$ (call ?model-obj randomInt ?*min-shack-rent* ?*max-shack-rent*)))  ;; rent is 200-500 R per month
            (tickStamp (current-tick))))
	(dump (str-cat ?hh " advertises a shack for rent"))
)

(set-current-module MAIN)