package org.cfpm.caves.za;

import jess.Context;
import jess.JessException;
import jess.RU;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;

/** This class implements a user function in Jess.
 * The function "number-of-common-attributes" expects two lists of
 * attributes and counts how many attributes are contained in both
 * lists. It is used to compare the tags of two agents to find out
 * how similar they are.
 * 
 * @author Scott
 *
 */
public class CommonAttributeCounter implements Userfunction {

	/** returns the name of the function in Jess */
	public String getName() {
		return "number-of-common-attributes";
	}

	/** calls the function with the given parameter and context
	 * from Jess.
	 * @param vv the list of parameters 
	 * @param c the Jess context
	 * @return the number of common attributes
	 */
	public Value call(ValueVector vv, Context c) throws JessException {
		ValueVector list1 = vv.get(1).listValue(c);
		ValueVector list2 = vv.get(2).listValue(c);
		int counter = 0;
		for (int i = 0; i < list1.size(); i++) {
			if (list1.get(i).intValue(c) == list2.get(i).intValue(c))
				counter++;
		}
		return new Value(counter, RU.INTEGER);
	}

}
