package org.cfpm.caves.za;

import jess.Context;
import jess.JessException;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;
import jess.RU;

/** This class implements a user function in Jess.
 * The parameterless function "current-tick" returns the current
 * model tick, to be used on the RHS of rules. 
 * This way is more efficient than having the model
 * in the LHS.
 * 
 * @author Ruth
 *
 */
public class CurrentTick implements Userfunction {

	/** a reference to the model */
	SouthAfricaModel model;
	
	/** Creates the user function with the given model
	 * 
	 * @param model the model
	 */
	public CurrentTick(SouthAfricaModel model){
		this.model = model;
	}
	
	/** returns the name of the function in Jess */	
	public String getName() {
		return "current-tick";
	}

	/** calls the function with the given parameter and context
	 * from Jess.
	 * @param vv the list of parameters (ignored) 
	 * @param c the Jess context (ignored)
	 * @return the current model tick
	 */
	public Value call(ValueVector vv, Context c) throws JessException {
		return new Value(model.getTick(), RU.INTEGER);
	}

}
