package org.cfpm.caves.za;

import jess.Context;
import jess.JessException;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;

/** This class implements a user function in Jess.
 * The parameterless function "dump" calls the method of the same
 * name on the model class. To be used on the RHS of rules.
 * This allows 
 * (a) to toggle output on/off via one model parameter, given
 * that all Jess output uses the dump function, and 
 * (b) an efficient way to call a method on the model without
 * having to put the model fact in the LHS of a rule.
 * 
 * @author Ruth
 *
 */
public class Dump implements Userfunction {
	
	/** a reference to the model */
	SouthAfricaModel model;
	
	/** Creates the user function with the given model
	 * 
	 * @param model the model
	 */
	public Dump(SouthAfricaModel model){
		this.model = model;
	}

	/** returns the name of the function in Jess */	
	public String getName() {
		return "dump";
	}

	/** calls the function with the given parameter and context
	 * from Jess.
	 * @param vv the list of parameters (ignored) 
	 * @param context the Jess context (ignored)
	 * @return NIL
	 */
	public Value call(ValueVector vv, Context context) throws JessException {
		String text = vv.get(1).stringValue(context);
		model.dump(text);
		return jess.Funcall.NIL;
	}

}
