package org.cfpm.caves.za;

import jess.Context;
import jess.JessException;
import jess.RU;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;
import java.util.HashMap;
import java.util.Iterator;

/** This userfunction returns the element with the highest number of occurrences in a list.
 * @author Ruth
 *
 */
public class ElementWithHighestOccurrence implements Userfunction {

	/** returns the name of the function in Jess */	
	public String getName() {
		return "element-with-highest-occurrence";
	}

	/** calls the function with the given parameter and context
	 * from Jess. This function has one parameter (a list of symbols) and returns
	 * the symbol with the highest number of occurrences 
	 * @param vv the list of parameters 
	 * @param c the Jess context 
	 * @return the symbol with highest number of occurrences
	 */
	public Value call(ValueVector vv, Context c) throws JessException {
		ValueVector list = vv.get(1).listValue(c);
		HashMap<String, Integer> occurrences = new HashMap<String, Integer>();
		// iterate over list and calculate number of occurrences of each distinct element
		for (int i = 0; i < list.size(); i++){
			String element = list.get(i).toString();
			if (occurrences.containsKey(element)){
				occurrences.put(element, occurrences.get(element) + 1);
			}
			else {
				occurrences.put(element, 1);
			}
		}
		// find element with highest number of occurrences
		String element = null;
		int max = 0;
		for (Iterator<String> it = occurrences.keySet().iterator(); it.hasNext(); ){
			String el = it.next();
			int occ = occurrences.get(el);
			if (occ > max){
				max = occ;
				element = el;
			}
		}
		return new Value(element, RU.SYMBOL);		
	}

}
