package org.cfpm.caves.za;

import jess.*;

/** This class implements a user function in Jess.
 * The function "endorsement-value" expects an endorsement scheme
 * and a list of endorsement as parameters. It returns the overall
 * endorsement value.
 * 
 * @author Scott
 *
 */
public class EndorsementValue implements Userfunction {
	
	/** returns the name of the function in Jess */	
	public String getName(){
		return "endorsement-value";
	}
	
	/**
	 * The first argument of the call is the endorsement scheme used
	 * to evaluate the endorsed object and the second is the list of
	 * endorsements on the object.
	 * 
	 * @see jess.Userfunction#call(jess.ValueVector, jess.Context)
	 * @param vv -- the ValueVector containing the header and parameters
	 * @param context -- the context of the Rete engine
	 * @return the overall endorsement value of the endorsed object
	 */
	public Value call(ValueVector vv, Context context) throws JessException{
		EndorsementScheme scheme = (EndorsementScheme) vv.get(1).javaObjectValue(context);
		ValueVector endorsementTokens = vv.get(2).listValue(context);
		double endValue = scheme.calculateEndorsementValue(endorsementTokens);
		Value result = new Value(endValue, RU.FLOAT);
		return result;
	}

}
