package org.cfpm.caves.za;

import jess.Context;
import jess.Userfunction;
import jess.JessException;
import jess.Value;
import jess.ValueVector;
import jess.Fact;


/** This class implements a user function in Jess.
 *  The function "assert-message-as-fact" can be used to assert the fact
 *  that is conveyed in a message.
 *  
 * @see template message in deftemplates.clp
 * @author Scott
 *
 */
public class MessageToFact implements Userfunction {
	
	/**
	 * The argument of the call is a message fact. This function "extracts" the fact
	 * conveyed in the message (slots header, slot-names, slot-values) and asserts it.
	 *
	 * @see jess.Userfunction#call(jess.ValueVector, jess.Context)
	 * @param vv -- the ValueVector containing the message fact (parameter to the user function)
	 * @param context -- the context of the Rete engine
	 */
	public Value call(ValueVector vv, Context context) throws JessException {
		Fact messageFact = vv.get(1).factValue(context);
		if (!messageFact.getName().equals("MAIN::message")){
			System.err.println("The fact should have header \"MAIN::message\" but instead has " + messageFact.getName());
			System.exit(1);
		}
		Fact f = new Fact(messageFact.getSlotValue("header").symbolValue(context), context.getEngine());
		ValueVector slotNames = messageFact.getSlotValue("slot-names").listValue(context);
		ValueVector slotValues = messageFact.getSlotValue("slot-values").listValue(context);
		for (int i = 0; i < slotNames.size(); i++){
			String slotName = slotNames.get(i).symbolValue(context);
			Value slotValue = slotValues.get(i);
			f.setSlotValue(slotName, slotValue);
		}
		context.getEngine().assertFact(f);
		return new Value(true);
	}

	/** returns the name of the function in Jess */	
	public String getName() {
		return "assert-message-as-fact";
	}

}
