package org.cfpm.caves.za;

import jess.Context;
import jess.FuncallTest;
import jess.JessException;
import jess.RU;
import jess.Userfunction;
import jess.Value;
import jess.ValueVector;
import jess.Funcall;

/** Calls the probabilityp method on the model class. For use in Jess on LHS of rules
 * with the test conditional element.
 * @author Ruth
 *
 */
public class ProbabilityP implements Userfunction {

	/** a reference to the model */
	SouthAfricaModel model;
	
	/** Creates the user function with the given model
	 * 
	 * @param model the model
	 */
	public ProbabilityP(SouthAfricaModel model){
		this.model = model;
	}

	/** returns the name of the function in Jess */	
	public String getName() {
		// TODO Auto-generated method stub
		return "probability-p";
	}

	/** returns true with the given probability p (calls the method probabilityp of the model).
	 * @param vv the list of parameters (in this case, a probability <= 1.0
	 * @param context the Jess context
	 * @return true if the random number sampled from a uniform distribution is lower than p,
	 * 		otherwise false
	 */
	public Value call(ValueVector vv, Context context) throws JessException {
		double p = Double.valueOf(vv.get(1).stringValue(context)).doubleValue();
		if (model.probabilityp(p))
			return Funcall.TRUE;
		return Funcall.FALSE;
	}

}
