package org.cfpm.caves.za;


import jess.*;
import java.util.ArrayList;

/** This class is used to import files into Jess at the beginning of 
 *  a simulation run.
 * 
 * @author Scott Moss, Ruth Meyer
 */
public class RulebaseCreator {
	
	/** a reference to the model */
	SouthAfricaModel model;
	/** a reference to the Jess engine */
	Rete engine;
	
	/** Creates a new rule base creator for the given model.
	 * 
	 * @param model the model
	 */
	public RulebaseCreator(SouthAfricaModel model){
		this.model = model;
		engine = model.engine;
	}
	
	/** initialises Jess by (a) loading user functions which need the model as parameter
	 * and (b) loading the template and rule definitions (.clp files)
	 */
	protected void initialise(){
		try {
			// load user functions
			engine.addUserfunction(new Dump(model));
			engine.addUserfunction(new CurrentTick(model));
			engine.addUserfunction(new ProbabilityP(model));
			// load template and query definitions 
			engine.batch("deftemplates.clp");
			engine.batch("defqueries.clp");
			// define global variables from model parameters
			// for faster access from Jess
			ArrayList<Defglobal> globalVars = new ArrayList<Defglobal>();
			globalVars.add(new Defglobal("*standard-wage*", new Value(model.standardWage, RU.FLOAT)));
			globalVars.add(new Defglobal("*similarity-age-range*", new Value(model.similarityAgeRange, RU.FLOAT)));
			globalVars.add(new Defglobal("*burial-cost*", new Value(model.burialCost, RU.FLOAT)));
			globalVars.add(new Defglobal("*random-partner-probability*", new Value(model.randomPartnerProbability, RU.FLOAT)));
			globalVars.add(new Defglobal("*shack-proportion*", new Value(model.shackProportion, RU.FLOAT)));
			globalVars.add(new Defglobal("*min-shack-rent*", new Value(model.minShackRent, RU.INTEGER)));
			globalVars.add(new Defglobal("*max-shack-rent*", new Value(model.maxShackRent, RU.INTEGER)));
			globalVars.add(new Defglobal("*house-cost*", new Value(model.houseCost, RU.INTEGER)));
			globalVars.add(new Defglobal("*min-remittance*", new Value(model.minRemittance, RU.INTEGER)));
			globalVars.add(new Defglobal("*max-remittance*", new Value(model.maxRemittance, RU.INTEGER)));
			globalVars.add(new Defglobal("*remittance-probability*", new Value(model.remittanceProbability, RU.FLOAT)));
			globalVars.add(new Defglobal("*school-fee*", new Value(model.schoolFee, RU.FLOAT)));
			globalVars.add(new Defglobal("*college-fee*", new Value(model.collegeFee, RU.FLOAT)));
			globalVars.add(new Defglobal("*min-lobola*", new Value(model.minLobola, RU.INTEGER)));
			globalVars.add(new Defglobal("*max-lobola*", new Value(model.maxLobola, RU.INTEGER)));
			globalVars.add(new Defglobal("*house-cost*", new Value(model.houseCost, RU.INTEGER)));
			engine.addDefglobals(globalVars);
			// load rules
			engine.batch("init.clp");
			engine.batch("content.clp");
			engine.batch("final.clp");
			engine.batch("cleanup.clp");	// most of this is now done via the disposal policy
		}
		catch(JessException je) {
			System.err.println("An error occurred at line " + je.getLineNumber());
			System.err.println("caused by " + je.getCause());
			System.err.println("Message: " + je.getDetail());
			je.printStackTrace();
		}	
	}
}
