
package org.cfpm.caves.za.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;

/** Splits the huge file containing all sexual networks into lots of files containing
 * exactly one sexual network. The files are named "sn-xxx.net" with xxx denoting the
 * tick they belong to.
 * @author Ruth
 *
 */
public class NetworkFileSplitter {
	
	static String inputPath = "/data/experiments/southAfricaModel/run-1202941021944/";
	static String outputPath = inputPath + "sexualNetworks/";
	static String inputFileName = "SexualNetwork.txt";
	static String outputFileName = "sn-";

	public static void splitNetworkFile(File networkFile) throws IOException {
		System.out.println("start processing file " + networkFile.getName());
		// open file input stream
		FileInputStream fileIn = new FileInputStream(networkFile);
		BufferedReader bufferIn = new BufferedReader(new InputStreamReader(fileIn));
		// create output directory
		File temp = new File(outputPath);
		if (!temp.exists()){
			if (temp.mkdir()){
				System.out.println("OUTPUT directory created");
			}
			else {
				System.out.println("FAILED to create OUTPUT directory");
			}
		}
		else {
			System.out.println("OUTPUT directory exists already");
		}
		// read in big network file and split it into separate files
		
		// ignore header with parameter settings --> read lines until you find a line that's just a 0
		String[] token = bufferIn.readLine().split("\\s");
		while (token != null && token.length > 0 && !token[0].equals("0")){
			token = bufferIn.readLine().split("\\s");
		}
		// now we've found the start of the first sexual network (for tick 0)
		
		while (bufferIn.ready() && token != null && token.length > 0) {
			// create output file
			File outputFile = new File(outputPath + outputFileName + token[0] + ".net");
			PrintWriter printOut = new PrintWriter(new FileWriter(outputFile));
			System.out.println("write network file " + outputFile.getName());
			// write lines to output file until there's a line with just one element that's not "*Edges" (= next tick)
			String fileLine = bufferIn.readLine();
			token = fileLine.split("\\s");
			while (token != null && (token.length > 1 || token[0].equals("*Edges"))){
				printOut.println(fileLine);
				fileLine = bufferIn.readLine();
				if (fileLine != null){
					token = fileLine.split("\\s");
				}
				else {
					token = null;
				}
			}
			printOut.flush();
			printOut.close();
		}
		
		// close file input stream
		bufferIn.close();
		fileIn.close();
		System.out.println("done");
	}
	
	public static void main(String[] args) {
		// args ignored for now, could write a version where input/output paths and files can be specified
		// as parameters
		File in = new File(inputPath + inputFileName);
		try {
			splitNetworkFile(in);
		}
		catch (IOException e){
			System.out.println("IOException!");
			e.printStackTrace();
		}
	}

}
